//+------------------------------------------------------------------+
//|                                                      csmooth.mqh |
//|                        Copyright 2014, MetaQuotes Software Corp. |
//|                                              http://www.mql5.com |
//+------------------------------------------------------------------+
#property copyright "Copyright 2019, rip"
#property link      "rip@yukr.net"
#property version   "1.00"

#include "json.mqh"

//+------------------------------------------------------------------+
//| Websocket class                                                  |
//+------------------------------------------------------------------+
class Websocket {
  private:
    string            url;
    string            host;
    int               port;
    int               socket;
    bool              ExtTLS;
    uint              timeout;
    bool              DEBUG;

    static const string GUID;
    static const uint key_len;

    static const uchar OPCODE_CONT;
    static const uchar OPCODE_TEXT;
    static const uchar OPCODE_BINARY;
    static const uchar OPCODE_CLOSE;
    static const uchar OPCODE_PING;
    static const uchar OPCODE_PONG;

    CJAVal            HTTPGetHeader(string header);
    string            DecodeBase64(string fData);
    string            EncodeBase64(string fData);
    bool              HTTPSend(string request);
    string            HTTPRecv();

    bool              Handshake();
    string            HASHCalculate(string skey);
    string            KEYGeneration();

    string            Unpack(uchar &data[], int len);

    string            Byte2BinaryStr(uchar data);
    string            Byte2BinaryStr(uchar &data[]);
    uint              Byte2Decimal(uchar &data[]);
    uint              Byte2Decimal(uchar data);
    int               Decimal2Byte(long data, uchar &data_out[], int in_rank);

    string            ValidateOPCODE(uchar &data[]);
    bool              ValidateFIN(uchar &data[]);
    bool              ValidateMASK(uchar &data[]);
    int               GetMSGLength(uchar &data[], uint &pos, int len);
    void              BuildMask(uchar &data_out[]);
    void              ApplayMask(uchar &mask[], uchar &payload[], uchar &payload_out[]);

  public:
    void              Websocket(bool debug);
    void              SetDefaultValues();
    void             ~Websocket();
    bool              Connect();
    void              Disconnect();
    bool              Connected();
    void              SetUrl(string in_url);
    string            GetUrl();
    void              SetHost(string in_host);
    void              SetPort(int in_port);

    string            Recv();
    bool              Send(string request);

    int               Pack(string data_src, uchar &data_dst[]);
};

static const string Websocket::GUID = "258EAFA5-E914-47DA-95CA-C5AB0DC85B11";
static const uint Websocket::key_len = 12;

static const uchar Websocket::OPCODE_CONT = 0;
static const uchar Websocket::OPCODE_TEXT = 1;
static const uchar Websocket::OPCODE_BINARY = 2;
static const uchar Websocket::OPCODE_CLOSE = 8;
static const uchar Websocket::OPCODE_PING = 9;
static const uchar Websocket::OPCODE_PONG = 10;


//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void Websocket::SetDefaultValues(void) {
    ExtTLS = false;
    timeout = 1000;
    host = "192.168.122.102";
    port = 5000;
    DEBUG = true;
}

//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void Websocket::Websocket(bool debug = false) {
    SetDefaultValues();
    DEBUG = debug;
    socket = SocketCreate();
    if(socket == INVALID_HANDLE) {
        Print("Websocket::Websocket: Socket creation error ", GetLastError());
    } else {
        if (DEBUG) Print("Websocket::Websocket: Socket created.");
    }
}


//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void Websocket::~Websocket() {
    Disconnect();
}

//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
bool Websocket::Connect() {
    if(socket != INVALID_HANDLE) {
        if(SocketConnect(socket, host, port, timeout)) {
            Print("Connected to ", host, ":", port);
            return Handshake();
        } else {
            Print("Connection ERROR. Сannot open connection to host: ", host, ":", port, " Error: ", GetLastError());
        }
    } else {
        Print("Socket creation error. Error: ", GetLastError());
    }
    return false;
}

//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
bool Websocket::Connected() {
    return SocketIsConnected(socket);
}

//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void Websocket::Disconnect() {
    if(SocketIsConnected(socket)) {
        if(!SocketClose(socket)) {
            Print("Возникла ошибка при закрытии сокета: ", GetLastError());
        }
    }
}

//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void Websocket::SetUrl(string in_url) {
    url = in_url;
}

//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void Websocket::SetHost(string in_host) {
    host = in_host;
}

//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void Websocket::SetPort(int in_port) {
    port = in_port;
}

//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
string Websocket::GetUrl(void) {
    return url;
}

//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
bool Websocket::HTTPSend(string request) {
    char req[];
    int  len = StringToCharArray(request, req) - 1;
    if(len > 0) {
        if(ExtTLS) {
            return(SocketTlsSend(socket, req, len) == len);
        } else {
            return(SocketSend(socket, req, len) == len);
        }
    } else {
        return(false);
    }
}

//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
string Websocket::HTTPRecv() {
    char    rsp[];
    string  result;
    uint    timeout_check = GetTickCount() + timeout;
    do {
        uint len = SocketIsReadable(socket);
        if(len) {
            int rsp_len;
            if(ExtTLS) {
                rsp_len = SocketTlsRead(socket, rsp, len);
            } else {
                rsp_len = SocketRead(socket, rsp, len, timeout);
            }
            if(rsp_len > 0) {
                result += CharArrayToString(rsp, 0, rsp_len);
                int header_end = StringFind(result, "\r\n\r\n");
                if(header_end > 0) {
                    return(StringSubstr(result, 0, header_end));
                }
            }
        }
    } while(GetTickCount() < timeout_check && !IsStopped());
    return("");
}

//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
bool Websocket::Send(string request) {
    if (StringLen(request) == 0) {
        request = "Hello";
    }
    uchar req[];
    int len = Pack(request, req);
    int send_len = -1;
    if(len > 0) {
        if(ExtTLS) {
            send_len = SocketTlsSend(socket, req, len);
        } else {
            send_len = SocketSend(socket, req, len);
        }
        if (DEBUG) Print(StringFormat("Websocket::Send: (rsp_len, len): %d,%d", send_len, len));
    }
    if (send_len < 0) {
        Print("Ошибка отправки данных. Error: ", GetLastError());
        return(false);
    } else {
        return(true);
    }
}

//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
string Websocket::Recv() {
    uchar    rsp[], res[];
    string  result = "";
    uint    timeout_check = GetTickCount() + timeout;
    do {
        uint len = SocketIsReadable(socket);
        if(len) {
            int rsp_len;
            if(ExtTLS) {
                rsp_len = SocketTlsRead(socket, rsp, len);
            } else {
                rsp_len = SocketRead(socket, rsp, len, timeout);
            }
            if(rsp_len > 0) {
                if (DEBUG) Print(StringFormat("Websocket::Recv: (rsp_len, len): %d,%d", rsp_len, len));
                ArrayInsert(res, rsp, ArraySize(res), 0);
            }
        }
    } while(GetTickCount() < timeout_check && !IsStopped());
    if(ArraySize(res) > 0) {
        result = Unpack(res, ArraySize(res));
        if (DEBUG) Print("Websocket::Recv: (result): ", result);
    }
    return(result);
}

//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
string Websocket::Unpack(uchar &data[], int len) {
    string res = "";
    uchar header[14];
    ArrayCopy(header, data, 0, 0, 14);
    int FIN = ValidateFIN(header);
    int MASK = ValidateMASK(header);
    string OPCODE = ValidateOPCODE(header);
    uint dataPosition = 0;
    int dataLength = GetMSGLength(header, dataPosition, len);
    if (DEBUG) Print("Websocket::Unpack: (dataLength): ", dataLength);
    res = CharArrayToString(data, dataPosition, dataLength, CP_UTF8);
    int filehandle = FileOpen("fractals2.csv", FILE_READ | FILE_WRITE | FILE_TXT);
    if(filehandle != INVALID_HANDLE) {
        FileSeek(filehandle, 0, SEEK_END);
        FileWriteString(filehandle, TimeToString(TimeCurrent(), TIME_DATE | TIME_SECONDS) + " Data length: " + IntegerToString(dataLength) + "\r");
        FileWriteString(filehandle, TimeToString(TimeCurrent(), TIME_DATE | TIME_SECONDS) + " " + res + "\r");
        FileClose(filehandle);
    }
    if (DEBUG) Print("Websocket::Unpack: (Data): ", res);
    return res;
}


//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
int Websocket::Pack(string data_src, uchar &data_dst[]) {
    int res = -1;
    uchar payload[];
    uchar header[];
    ArrayResize(header, 2);
    ArrayInitialize(header, 0);
    header[0] = header[0] | 128;     //OP_TEXT
    header[0] = header[0] | 1;       //FIN
    long payload_length = StringToCharArray(data_src, payload, 0, StringLen(data_src), CP_UTF8);
    if (payload_length <= 125) {
        header[1] = (uchar)payload_length;
    } else if (payload_length >= 126 && payload_length < 65535) {
        header[1] = 126;             //code = 126
        ArrayResize(header, ArraySize(header) + 2);
        uchar payload_length_byte[];
        int count = Decimal2Byte(payload_length, payload_length_byte, 2);
        ArrayCopy(header, payload_length_byte, 2, 0, count);
    } else if (payload_length >= 65535 && payload_length <= 18446744073709551615) {
        header[1] = 127;
        ArrayResize(header, ArraySize(header) + 8);
        uchar payload_length_byte[];
        int count = Decimal2Byte(payload_length, payload_length_byte, 8);
        ArrayCopy(header, payload_length_byte, 2, 0, count);
    } else {
        Print("Сообщение слишком большое. Разбейте его на куски");
        return(res);
    }
    header[1] = header[1] | 128;     //MASK
    ArrayResize(header, ArraySize(header) + 4);
    uchar mask[4];
    BuildMask(mask);
    ArrayCopy(header, mask, ArraySize(header) - 4, 0, 4);
    uchar payload_pack[];
    ApplayMask(mask, payload, payload_pack);
    ArrayResize(data_dst, ArraySize(payload_pack) + ArraySize(header));
    ArrayInitialize(data_dst, 0);
    ArrayCopy(data_dst, header, 0, 0, ArraySize(header));
    ArrayCopy(data_dst, payload_pack, ArraySize(header), 0, ArraySize(payload_pack));
    res = ArraySize(data_dst);
    return(res);
}
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void Websocket::BuildMask(uchar &data_out[]) {
    MathSrand(GetTickCount());
    if (ArraySize(data_out) == 0) {
        ArrayResize(data_out, 4);
    }
    ArrayInitialize(data_out, 0);
    for (int i = 0; i < 4; i++) {
        int rnd = MathRand();
        data_out[i] = (uchar)rnd;
    }
}

//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void Websocket::ApplayMask(uchar &mask[], uchar &payload[], uchar &payload_out[]) {
    int i_index = 0;
    ArrayResize(payload_out, ArraySize(payload));
    ArrayCopy(payload_out, payload, 0, 0, ArraySize(payload));
    do {
        for (int j_index = 0; j_index < ArraySize(mask); j_index++) {
            payload_out[i_index] ^= mask[j_index];
            i_index++;
            if(i_index >= ArraySize(payload_out)) {
                break;
            }
        }
    } while(i_index < ArraySize(payload_out));
}


//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
string Websocket::Byte2BinaryStr(uchar &data[]) {
    string res = "";
    for (int i = 0; i < ArraySize(data); i++) {
        res += Byte2BinaryStr(data[i]);
    }
    return(res);
}

//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
string Websocket::Byte2BinaryStr(uchar data) {
    uchar num = data;
    uchar res[8];
    ArrayInitialize(res, (uchar)StringGetCharacter("0", 0));
    int i_index = 8;
    do {
        i_index--;
        double t = MathMod(num, 2);
        if(t == 1) {
            res[i_index] = (uchar)StringGetCharacter("1", 0);
        } else {
            res[i_index] = (uchar)StringGetCharacter("0", 0);
        }
        num = (uchar)MathRound(num / 2);
    } while(num > 0);
    return(CharArrayToString(res));
}
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
int Websocket::Decimal2Byte(long data, uchar &data_out[], int in_rank) {
    long num = data;
    //if (DEBUG) Print("Websocket::Decimal2Byte: (byte_rank): ", byte_rank);
    int i_index = (int)in_rank - 1;
    int j_index = (int)in_rank * 8;
    ArrayResize(data_out, in_rank);
    ArrayInitialize(data_out, 0);
    double t = 0;
    do {
        if (num > 0) {
            t = MathMod(num, 2);
            num = (long)MathRound(num / 2);
        } else {
            t = 0;
        }
        j_index--;
        if(t == 1) {
            data_out[i_index] = data_out[i_index] >> 1;
            data_out[i_index] = data_out[i_index] | 128;
        } else {
            data_out[i_index] = data_out[i_index] >> 1;
        }
        double n = MathMod(j_index, 8);
        if (n == 0) {
            i_index--;
        }
    } while(j_index > 0);
    return ArraySize(data_out);
}


//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
uint Websocket::Byte2Decimal(uchar &data[]) {
    uint num = data[ArraySize(data) - 1];
    for(int i = ArraySize(data) - 2; i >= 0; i--) {
        num = num | (data[0] << (8 * (i + 1)));
    }
    return num;
}

//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
uint Websocket::Byte2Decimal(uchar data) {
    uint res = 0;
    int j_index = 0;
    uchar num = data;
    do {
        res += (uint)MathMod(num, 2) * (uint)MathPow(2, j_index);
        j_index++;
        num = (uchar)MathRound(num / 2);
    } while(num > 0);
    return res;
}


//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
int Websocket::GetMSGLength(uchar &data[], uint &pos, int len) {
    int msg_length = -1;
    uchar res = data[1] & 127;
    pos = ValidateMASK(data) * 4;
    if(res <= 125) {
        msg_length = (int)Byte2Decimal(res);
        pos += 2;
    } else if(res == 126) {
        uchar t[];
        ArrayCopy(t, data, 0, 2, 2);
        msg_length = (int)Byte2Decimal(t);
        pos += 4;
    } else if(res == 127) {
        pos += 10;
        uchar t[];
        ArrayCopy(t, data, 0, 2, 8);
        msg_length = (int)Byte2Decimal(t);
    }
    return(msg_length);
}

//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
bool Websocket::ValidateFIN(uchar &data[]) {
    uchar res = data[0] & 128;
    if(res == 0) {
        return(false);
    } else {
        return(true);
    }
}

//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
bool Websocket::ValidateMASK(uchar &data[]) {
    uchar res = data[1] & 128;
    if(res == 0) {
        return(false);
    } else {
        return(true);
    }
}

//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
string Websocket::ValidateOPCODE(uchar &data[]) {
    uchar res = data[0] & 15;
    if(res == OPCODE_TEXT) {
        return("TEXT");
    } else if(res == OPCODE_BINARY) {
        return("BINARY");
    } else if(res == OPCODE_CLOSE) {
        return("CLOSE");
    } else if(res == OPCODE_CONT) {
        return("CONT");
    } else if(res == OPCODE_PING) {
        return("PING");
    } else if(res == OPCODE_PONG) {
        return("PONG");
    } else {
        return(Byte2BinaryStr(res));
    }
}

//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
bool Websocket::Handshake() {
    string key = KEYGeneration();
    string   request = "GET / HTTP/1.1 101 Switching Protocols\r\n";
    request += "host: " + host + ":5000\r\n";
    request += "origin: http://" + host + ":5000\r\n";
    request += "upgrade: websocket\r\n";
    request += "connection: upgrade\r\n";
    request += "sec-websocket-Key:" + key + "\r\n";
    request += "sec-websocket-version: 13\r\n";
    request += "\r\n";
    if(HTTPSend(request)) {
        string res = HTTPRecv();
        if(StringLen(res) > 0) {
            CJAVal header = HTTPGetHeader(res);
            if(header["sec-websocket-accept"] != HASHCalculate(key)) {
                Print("Полученные ключ не совпадает с отправленным.");
                //Disconnect();
                return (false);
            } else {
                return (true);
            }
        } else {
            Print("Не удалось получить ответ, ошибка ", GetLastError());
            return (false);
        }
    } else {
        Print("Не удалось отправить GET-запрос, ошибка ", GetLastError());
        return (false);
    }
}

//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
string Websocket::KEYGeneration() {
    uchar res[], src[], key[];
    int min = 0;
    int max = 1024;
    uchar rnd_char;
    ArrayResize(res, key_len);
    string res_str = "";
    for(uint i = 0; i < key_len; i++) {
        rnd_char = (uchar)(min + (int)((MathRand() / 32768.0) * (max - min)));
        res_str += StringFormat("%.2x", rnd_char);
    }
    return res_str;
}


//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
string Websocket::EncodeBase64(string fData) {
    uchar data[], key[], result[];
    return !StringToCharArray(fData, data)              ||
           !CryptEncode(CRYPT_BASE64, data, key, result)  ? NULL : CharArrayToString(result);
}

//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
string Websocket::DecodeBase64(string fData) {
    uchar data[], key[], result[];
    return !StringToCharArray(fData, data)              ||
           !CryptDecode(CRYPT_BASE64, data, key, result)  ? NULL : CharArrayToString(result);
}

//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
string Websocket::HASHCalculate(string skey) {
    uchar data[], key[], result[], dst[];
    return !StringToCharArray(skey + GUID, data, 0, StringLen(skey + GUID), CP_UTF8) ||
           !CryptEncode(CRYPT_HASH_SHA1, data, key, dst) || !CryptEncode(CRYPT_BASE64, dst, key, result) ? NULL : CharArrayToString(result);
}

//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
CJAVal Websocket::HTTPGetHeader(string header) {
    CJAVal res;
    string result[];
    ushort endstr_sep = 10;
    ushort colon_sep = StringGetCharacter(":", 0);
    int k = StringSplit(header, endstr_sep, result);
    for(int i = 0; i < ArraySize(result); i++) {
        StringTrimRight(result[i]);
        StringTrimLeft(result[i]);
        string dict[];
        k = StringSplit(result[i], colon_sep, dict);
        if(k > 1) {
            StringTrimRight(dict[0]);
            StringToLower(dict[0]);
            StringTrimLeft(dict[1]);
            res[dict[0]] = dict[1];
        }
    }
    return res;
}
//+------------------------------------------------------------------+
